local function split(str, separator)
   local sep, ret = separator or "|", {}
   local pattern = string.format("([^%s]+)", sep)
   string.gsub(str, pattern, function(m) ret[#ret+1] = m end)
   return ret
end

function clubs(label, rclubs, lclubs, start, delay)
   local rprint, lprint, delaylabel
   rprint = 'R' .. rclubs
   lprint = 'L' .. lclubs
   delaylabel = ''
   if delay then
      if string.len(delay) == 2 then
         delaylabel = '$+\\frac' .. delay .. '$'
      else
         delaylabel = '$+' .. delay .. '$'
      end
   end
   if start == 'r' then
      rprint = 'R' .. rclubs .. delaylabel
   elseif start == 'lr' or start == 'rl' then
      rprint = 'R' .. rclubs .. delaylabel
      lprint = 'L' .. lclubs .. delaylabel
   elseif start == 'l' then
      -- reverse positions
      lprint = rprint
      rprint = 'L' .. lclubs .. delaylabel
   end
   tex.print('$' .. label .. '$: ' .. rprint .. '/' .. lprint)
end

function juggler(x, y, rot, label, rclubs, lclubs, start, delay, style)
   local offset, length, xreference, yreference, xstart, ystart, xend, yend
   local xx, yy, delayrotate, delaylabel
   label = label or ''
   rclubs = rclubs or 0
   lclubs = lclubs or 0
   if style then
      style = '[' .. style .. ']'
   else
      style = ''
   end
   tex.print('\\begin{scope}[shift={(' .. x .. ',' .. y ..')},rotate=' .. rot
             .. ']')
   tex.print('\\filldraw[fill=white] ' .. style
             .. ' (2,0) to (1,1) to[out=135, in=45] (-1,1) '
             .. 'to[out=225,in=135] (-1,-1) to[out=315,in=225] (1,-1) to cycle;')
   offset = .33
   length = .8
   xreference = 2 + offset/4
   yreference = offset/4
   for i = 1, rclubs do
      xstart = xreference - offset*i
      ystart = yreference - offset*i
      xend = xreference + length - offset*i
      yend = yreference - length - offset*i
      tex.print('\\draw [thick] (' .. xstart .. ',' .. ystart .. ') -- ('
                .. xend .. ',' .. yend .. ');')
   end
   for i = 1, lclubs do
      xstart = xreference - offset*i
      ystart = - yreference + offset*i
      xend = xreference + length - offset*i
      yend = - yreference + length + offset*i
      tex.print('\\draw [thick] (' .. xstart .. ',' .. ystart .. ') -- ('
                .. xend .. ',' .. yend .. ');')
   end
   if start == 'r' then
      xx = xreference + length - offset
      yy = yreference - length - offset
      tex.print('\\draw [fill] (' .. xx .. ',' .. yy .. ') circle [radius=.2];')
   elseif start == 'l' then
      xx = xreference + length - offset
      yy = - yreference + length + offset
      tex.print('\\draw [fill] (' .. xx .. ',' .. yy .. ') circle [radius=.2];')
   elseif start == 'rl' or start == 'lr'  then
      xx = xreference + length - offset
      yy = - yreference + length + offset
      tex.print('\\draw [fill] (' .. xx .. ',' .. yy .. ') circle [radius=.2];')
      xx = xreference + length - offset
      yy = yreference - length - offset
      tex.print('\\draw [fill] (' .. xx .. ',' .. yy .. ') circle [radius=.2];')
   end
   if delay and start then
      if string.len(delay) == 2 then
         delaylabel = '\\scalebox{.66}{$+\\mathrlap{\\frac' .. delay .. '}$}'
      else
         delaylabel = '\\scalebox{.66}{$+\\mathrlap{' .. delay .. '}$}'
      end
      if start == 'r' or start == 'rl' then
         yy = yy - .8
         delayrotate = '270'
      else
         yy = yy + .8
         delayrotate = '90'
      end
      tex.print('\\node [rotate=' .. delayrotate .. ', transform shape] at ('
                .. xx .. ',' .. yy .. ') {' .. delaylabel .. '};')
   end
   tex.print('\\end{scope}')
   tex.print('\\node at (' .. x .. ',' .. y ..') {$' .. label .. '$};')
end

function ghost(x, y, rot, label, rclubs, lclubs)
   juggler(x, y, rot, label, rclubs, lclubs, nil, nil, 'dashed')
end

function magify(x, y, rot, side, num)
   tex.print('\\begin{scope}[shift={(' .. x .. ',' .. y ..')},rotate=' .. rot
             .. ']')
   offset = .33
   length = .8
   xreference = 2 + offset/4
   yreference = offset/4
   if side == 'r' then
      mult = 1
   else
      mult = -1
   end
   xstart = xreference - offset*num
   ystart = mult*(yreference - offset*num)
   xend = xreference + length - offset*num
   yend = mult*(yreference - length - offset*num)
   tex.print('\\draw [thick, dash pattern=on .7pt off .7pt, color=lightgray] (' .. xstart .. ',' .. ystart .. ') -- ('
             .. xend .. ',' .. yend .. ');')
   tex.print('\\end{scope}')
end

local function _find_target(nodes, juggler, index, thrownum, receiver)
   local possible_receivers, eta, candidate, score, fit
   if receiver then
      possible_receivers = {receiver}
   else
      possible_receivers = {}
      for i = 1,#nodes do
         if juggler + i <= #nodes then
            table.insert(possible_receivers, juggler + i)
         end
         if juggler - i > 0 then
            table.insert(possible_receivers, juggler - i)
         end
      end
   end
   eta = (nodes[juggler][index].time
          + nodes[juggler][index].throws[thrownum].offset)
   candidate = nil
   score = nil
   for _, possible_receiver in ipairs(possible_receivers) do
      for j, node in ipairs(nodes[possible_receiver]) do
         fit = node.time - eta
         if (fit >= -.3 and fit <= .6 and node.reverse == nil and
             (score == nil or math.abs(fit) < score)) then
                candidate = {juggler=possible_receiver, index=j}
                score = math.abs(fit)
         end
      end
   end
   return candidate
end

function causaldia(spec)
   local lines, atoms, passoffset, jugglers, juggler, first, nodes, separators
   local nodelist, time, lastsymbol, nextsymbol, skips_by_separator, separator
   local tmp, bihandlist, bihandshift, shift, node, throwlist, throw, isself
   local zipmultiplier, reference, target, equals, arrows, ypos, colorlabel
   local dashlabel, angle, times, count, bendlabel, arrow, sepstyle, lower
   local dotlabel
   --
   -- 1. prepare the input
   --
   lines = split(spec, '|')
   atoms = {}
   for i, line in ipairs(lines) do
      atoms[i] = split(line, ' ')
   end
   passoffset = 0
   jugglers = {}
   for i, atomlist in ipairs(atoms) do
      juggler = string.char(64 + i)
      first = atomlist[1]
      if (string.sub(first, 1, 1) == '('
          and string.sub(first, -1, -1) == ')') then
         juggler = string.sub(first, 2, -2)
         table.remove(atomlist, 1)
      end
      jugglers[i] = juggler
      first = atomlist[1]
      if (string.sub(first, 1, 3) == '.5/'
          or string.sub(first, 2, 4) == '.5/') then
         passoffset = .5
      end
   end
   --
   -- 2. parse the nodes
   --
   nodes = {}
   separators = {}
   for i, atomlist in ipairs(atoms) do
      nodelist = {}
      time = -1
      lastsymbol = ''
      nextsymbol = 'R'
      skips_by_separator = 0
      for j, atom in ipairs(atomlist) do
         -- check for sole separators
         if atom == ';' or atom == '!' then
            separator = {left={juggler=i, index=j-1-skips_by_separator},
                         right={juggler=i, index=j-skips_by_separator}, style=nil}
            if atom == ';' then
               separator.style = 'dashed'
            end
            table.insert(separators, separator)
            skips_by_separator = skips_by_separator + 1
            goto skipseparator
         end
         -- now we have an honest node
         -- first check the timing
         tmp = split(atom, '/')
         if #tmp > 1 then
            time = time + tonumber(tmp[1])
            if j == 1 then
               time = time + 1
            end
            atom = tmp[2]
         else
            time = time + 1
         end
         -- second check for a separator
         first = string.sub(atom, 1, 1)
         if first == ';' or first == '!' then
            separator = {left={juggler=i, index=#nodelist + 1},
                         right={juggler=i, index=#nodelist + 1}, style=nil}
            if first == ';' then
               separator.style = 'dashed'
            end
            table.insert(separators, separator)
            atom = string.sub(atom, 2, -1)
         end
         -- third split for bihanded throws
         bihandlist = split(atom, '=')
         bihandshift = -.4
         shift = 0
         if #bihandlist > 1 then
            shift = -bihandshift/2
         end
         for _, biatom in ipairs(bihandlist) do
            node = {reverse=nil, time=time, shift=shift, isnone=false,
                    throws={}}
            -- fourth check for custom symbol
            tmp = split(biatom, '.')
            if #tmp > 1 then
               node.symbol = tmp[1]
               biatom = tmp[2]
            else
               node.symbol = nextsymbol
            end
            if node.symbol == 'R' then
               nextsymbol = 'L'
            elseif node.symbol == 'L' then
               nextsymbol = 'R'
            end
            if ((node.symbol == 'R' or node.symbol == 'L')
                and node.symbol == lastsymbol and #bihandlist == 1) then
               node.color = 'red'
            else
               node.color = 'black'
            end
            lastsymbol = node.symbol
            -- fifth check for none
            if biatom == 'N' then
               node.isnone = true
               goto skipnone
            end
            -- sixth split for multiplex
            throwlist = split(biatom, ',')
            for _, throwatom in ipairs(throwlist) do
               throw = {juggler=nil, index=nil, hint=nil, explicit=nil,
                        empty=nil}
               isself = false
               zipmultiplier = 1
               -- seventh check kind of throw
               first = string.sub(throwatom, 1, 1)
               if first == 'H' then
                  goto skiphold
               elseif first == 'S' or first == 'Z' or first == 'E' then
                  throw.juggler = i
                  isself = true
                  if first == 'Z' then
                     zipmultiplier = -1
                  elseif first == 'E' then
                     zipmultiplier = -1
                     throw.empty = true
                  end
               end
               throwatom = string.sub(throwatom, 2, -1)
               -- eigth determine offset
               first = string.match(throwatom, '^-?%d+')
               if first then
                  if isself then
                     throw.offset = zipmultiplier * tonumber(first)
                  else
                     throw.offset = tonumber(first) + passoffset
                  end
                  throwatom = string.sub(throwatom, #first + 1, -1)
               else
                  if isself then
                     throw.offset = zipmultiplier*1
                  else
                     throw.offset = 1 + passoffset
                  end
               end
               -- ninth check for explicit target
               if string.sub(throwatom, 1, 1) == '@' then
                  throwatom = string.sub(throwatom, 2, -1)
                  first = string.match(throwatom, '^%d+')
                  throw.juggler = tonumber(first)
                  throwatom = string.sub(throwatom, #first + 1, -1)
                  if (string.sub(throwatom, 1, 1) == '+'
                      or string.sub(throwatom, 1, 1) == '-') then
                     first = string.match(throwatom, '^[+-]%d+')
                     throw.explicit = tonumber(first)
                     throwatom = string.sub(throwatom, #first + 1, -1)
                  end
                  if string.sub(throwatom, 1, 1) == ':' then
                     throwatom = string.sub(throwatom, 2, -1)
                     first = string.match(throwatom, '^%d+')
                     throw.hint = tonumber(first)
                     throwatom = string.sub(throwatom, #first + 1, -1)
                  end
               end
               if string.len(throwatom) > 0 then
                  print('Warning: leftover specification after parsing "'
                        .. throwatom .. '" at (' .. i .. '/' .. j .. ')')
               end
               table.insert(node.throws, throw)
               ::skiphold::
            end
            ::skipnone::
            shift = shift + bihandshift
            table.insert(nodelist, node)
         end
         ::skipseparator::
      end
      table.insert(nodes, nodelist)
   end
   --
   -- 3. assign targets
   --
   -- first resolve explicits
   for i, nodelist_ in ipairs(nodes) do
      for j, node_ in ipairs(nodelist_) do
         for k, throw_ in ipairs(node_.throws) do
            if throw_.juggler ~= nil and throw_.explicit ~= nil then
               for l, other in ipairs(nodes[throw_.juggler]) do
                  if other.time >= node_.time then
                     reference = l
                     break
                  end
               end
               throw_.index = reference + throw_.explicit
               nodes[throw_.juggler][throw_.index].reverse = {
                   juggler=i, index=j, throw=k}
            end
         end
      end
   end
   -- second resolve hinted
   for i, nodelist_ in ipairs(nodes) do
      for j, node_ in ipairs(nodelist_) do
         for k, throw_ in ipairs(node_.throws) do
            if throw_.hint ~= nil and throw.index == nil then
               target = _find_target(nodes, i, j, k, throw_.juggler)
               if target then
                  time = nodes[target.juggler][target.index].time
                  equals = {}
                  for l, other in ipairs(nodes[target.juggler]) do
                     if other.time == time then
                        table.insert(equals, l)
                     end
                  end
                  throw_.juggler = target.juggler
                  throw_.index = equals[throw_.hint]
                  nodes[throw_.juggler][throw_.index].reverse = {
                      juggler=i, index=j, throw=k}
               end
            end
         end
      end
   end
   -- third resolve selves
   for i, nodelist_ in ipairs(nodes) do
      for j, node_ in ipairs(nodelist_) do
         for k, throw_ in ipairs(node_.throws) do
            if throw_.juggler == i and throw_.index == nil then
               target = _find_target(nodes, i, j, k, throw_.juggler)
               if target then
                  throw_.juggler = target.juggler
                  throw_.index = target.index
                  nodes[target.juggler][target.index].reverse = {
                      juggler=i, index=j, throw=k}
               end
            end
         end
      end
   end
   -- fourth resolve known passes
   for i, nodelist_ in ipairs(nodes) do
      for j, node_ in ipairs(nodelist_) do
         for k, throw_ in ipairs(node_.throws) do
            if throw_.juggler ~= nil and throw_.index == nil then
               target = _find_target(nodes, i, j, k, throw_.juggler)
               if target then
                  throw_.juggler = target.juggler
                  throw_.index = target.index
                  nodes[target.juggler][target.index].reverse = {
                      juggler=i, index=j, throw=k}
               end
            end
         end
      end
   end
   -- fifth resolve unknown passes
   for i, nodelist_ in ipairs(nodes) do
      for j, node_ in ipairs(nodelist_) do
         for k, throw_ in ipairs(node_.throws) do
            if throw_.index == nil then
               target = _find_target(nodes, i, j, k, throw_.juggler)
               if target then
                  throw_.juggler = target.juggler
                  throw_.index = target.index
                  nodes[target.juggler][target.index].reverse = {
                      juggler=i, index=j, throw=k}
               else
                  print('Warning: unresolved throw (' .. i .. '/' .. j .. '/'
                        .. k ..')')
                  print('part of: ' .. spec)
               end
            end
         end
      end
   end
   --
   -- 4. create output
   --
   arrows = {}
   for i, nodelist_ in ipairs(nodes) do
      tex.print('\\node at (-.5,' .. -i .. ') {$' .. jugglers[i] .. '$:};')
      for j, node_ in ipairs(nodelist_) do
         ypos = -i+node_.shift
         if node_.isnone then
            tex.print('\\node (' .. i .. '/' .. j .. ') at (' ..
                         node_.time .. ',' .. ypos .. ') {};')
         else
            tex.print('\\node (' .. i .. '/' .. j .. ') [' .. node_.color ..
                         '] at (' .. node_.time .. ',' .. ypos .. ') {' ..
                         node_.symbol .. '};')
            for _, throw_ in ipairs(node_.throws) do
               if throw_.index then
                  if (node_.symbol == nodes[throw_.juggler][throw_.index].symbol and
                      throw_.juggler ~= i) then
                     colorlabel = ', blue'
                  else
                     colorlabel = ''
                  end
                  if nodes[throw_.juggler][throw_.index].isnone then
                     dashlabel = ', dashed'
                  else
                     dashlabel = ''
                  end
                  if throw_.empty then
                     dotlabel = ', dotted'
                  else
                     dotlabel = ''
                  end
                  if throw_.juggler == i and math.abs(throw_.index - j) > 1 then
                     if i <= (#jugglers+1)/2 then
                        angle = 30
                     else
                        angle = -30
                     end
                     if throw_.index < j then
                        angle = 180 - angle
                     end
                     bendlabel = '[out=' .. angle .. ',in=' .. (180 - angle) .. ']'
                     if math.abs(throw_.index - j) < 4 and j < throw_.index then
                        times = {}
                        for l = j, throw_.index do
                           times[nodelist_[l].time] = true
                        end
                        count = 0
                        for _, _ in pairs(times) do
                           count = count + 1
                        end
                        if count < 3 then
                           bendlabel = ''
                        end
                     end
                  else
                     bendlabel = ''
                  end
                  arrow = ('\\draw [->' .. colorlabel .. dashlabel .. dotlabel
                           .. '] (' .. i .. '/' .. j .. ') to ' .. bendlabel
                           .. ' (' .. throw_.juggler .. '/' .. throw_.index
                           .. ');')
                  table.insert(arrows, arrow)
               end
            end
         end
      end
   end
   for _, arrow_ in ipairs(arrows) do
      tex.print(arrow_)
   end
   for _, separator_ in ipairs(separators) do
      time = nodes[separator_.left.juggler][separator_.left.index].time
      time = time + nodes[separator_.right.juggler][separator_.right.index].time
      time = time /2
      sepstyle = ''
      lower = - #jugglers - .5
      if separator_.style then
         sepstyle = '[' .. separator_.style .. ']'
      end
      tex.print('\\draw ' .. sepstyle .. ' (' .. time .. ', -.5) -- (' .. time .. ', ' .. lower .. ');')
   end
end

local function _str_to_safe_str(s)
   local ret = ''
   if s ~= nil then
     for i = 1, #s do
        local c = string.sub(s, i, i)
        local b = string.byte(c)
        if b >= 65 and b <= 90 then
           ret = ret .. 'l' .. c
        elseif b >= 97 and b <= 122 then
           ret = ret .. 'l' .. c
        elseif b >= 48 and b <= 57 then
           ret = ret .. 'd' .. string.char(b + 20)
        elseif c == '_' then
           ret = ret .. 'su'
        elseif c == '-' then
           ret = ret .. 'sm'
        elseif c == '.' then
           ret = ret .. 'sd'
        elseif c == ',' then
           ret = ret .. 'sc'
        else
           ret = ret ..'xx'
        end
     end
   else
      ret = 'undefined'
   end
   return ret
end

local function _transition_sequence(jugglers, nodes)
   local transition_matrix, reverse_jugglers, columns, tm, node, valid, current
   local output, ret, result_matrix, target
   valid = true
   transition_matrix = {}
   for i = 1, #jugglers do
      transition_matrix[i] = i
   end
   result_matrix = {}
   reverse_jugglers = {}
   for k, v in pairs(jugglers) do
      reverse_jugglers[v] = k
   end
   columns = #nodes[1]
   for _, nodelist in pairs(nodes) do
      columns = math.min(columns, #nodelist)
   end
   for column = 1, columns do
      for juggler = 1, #nodes do
         node = nodes[juggler][column]
         if node.base == 'i' then
            target = reverse_jugglers[node.subscript]
            if target ~= nil then
               tm = transition_matrix
               tm[juggler], tm[target] = tm[target], tm[juggler]
            else
               valid = false
            end
         end
         -- arrow nodes are always the very last ones, so we can
         if node.arrow then
            target = reverse_jugglers[node.base]
            if target ~= nil then
               result_matrix[transition_matrix[juggler]] = target
            else
               valid = false
            end
         end
      end
   end
   ret = ''
   if valid then
      ret = ret .. '$'
       current = 1
       for _ = 1, #jugglers do
          ret = ret .. jugglers[current] .. '\\to '
          current = result_matrix[current]
       end
       ret = ret .. jugglers[1] .. '$'
       if #result_matrix ~= #jugglers then
          ret = '?????'
       end
   else
      print('Warning: invalid transition table')
      ret = '???'
   end
   return ret
end

function juggletable(spec)
   local lines, atoms, jugglers, juggler, first, nodes, nodelist, node
   local issuper, script, count, start, dotter, super, sub, hassuper
   --
   -- 1. prepare the input
   --
   lines = split(spec, '|')
   atoms = {}
   for i, line in ipairs(lines) do
      atoms[i] = split(line, ' ')
   end
   jugglers = {}
   for i, atomlist in ipairs(atoms) do
      juggler = string.char(64 + i)
      first = atomlist[1]
      if string.sub(first, 1, 1) == '(' and string.sub(first, -1, -1) == ')' then
         juggler = string.sub(first, 2, -2)
         table.remove(atomlist, 1)
      end
      jugglers[i] = juggler
   end
   --
   -- 2. parse the nodes
   --
   nodes = {}
   for i, atomlist in ipairs(atoms) do
      nodelist = {}
      for j, atom in ipairs(atomlist) do
         node = {empty=false, phantom=false, arrow=false, dots=0, base='', superscript='', subscript=''}
         -- check for sole separators
         if atom == '-' or atom == '=' then
            node.empty = true
            if atom == '=' then
               node.phantom = true
            end
            goto skipempty
         end
         if string.sub(atom, 1, 1) == '>' then
            node.arrow = true
            node.base = string.sub(atom, 2, -1)
            goto skiparrows
         end
         if string.sub(atom, 1, 1) == '{' and string.sub(atom, -1, -1) == '}' then
            node.base = atom
            goto skipraw
         end
         while string.sub(atom, 1, 1) == '.' do
            node.dots = node.dots + 1
            atom = string.sub(atom, 2, -1)
         end
         first = string.byte(string.sub(atom, 1, 1))
         if first >= 65 and first <= 90 then
            node.base = 'p'
            node.subscript = string.sub(atom, 1, 1)
            atom = string.sub(atom, 2, -1)
         elseif string.sub(atom, 1, 1) == 's' then
            node.base = '\\selfsym{}'
            atom = string.sub(atom, 2, -1)
         elseif first >= 97 and first <= 122 then
            node.base = string.sub(atom, 1, 1)
            atom = string.sub(atom, 2, -1)
         end
         while string.sub(atom, 1, 1) == '^' or string.sub(atom, 1, 1) == '_' do
            issuper = (string.sub(atom, 1, 1) == '^')
            atom = string.sub(atom, 2, -1)
            script = ''
            while (string.len(atom) > 0 and
                   (string.sub(atom, 1, 1) ~= '^' and string.sub(atom, 1, 1) ~= '_')) do
               script = script .. string.sub(atom, 1, 1)
               atom = string.sub(atom, 2, -1)
            end
            if issuper then
               node.superscript = node.superscript .. script
            else
               node.subscript = node.subscript .. script
            end
         end
         if ((node.dots > 0 or string.len(node.superscript) > 0 or string.len(node.subscript) > 0)
             and node.base == '\\selfsym{}') then
            node.base = 's'
         end
         if string.len(atom) > 0 then
            print('Warning: leftover specification after parsing "'
                  .. atom .. '" at (' .. i .. '/' .. j .. ')')
         end
         ::skipempty::
         ::skiparrows::
         ::skipraw::
         table.insert(nodelist, node)
      end
      table.insert(nodes, nodelist)
   end
   --
   -- 3. create output
   --
   --tex.print('\\rule{2pt}{' .. #nodes / 2 + .333 .. '\\baselineskip}%') -- a strut
   count = 0
   for _, nodelist_ in ipairs(nodes) do
      if #nodelist_ > count then
         count = #nodelist_
      end
   end
   count = count + 1 -- include names at the start
   start = '\\begin{tabular}{'
   for _ = 1, count - 1 do
      start = start .. 'c'
   end
   start = start .. 'l}' -- last column left aligned on the arrows
   tex.print(start)
   for i, nodelist_ in ipairs(nodes) do
      hassuper = false
      tex.print('$' .. jugglers[i] .. '$%')
      if i == 1 then
         tex.print('\\vphantom{\\scalebox{1.25}{M}}%') -- a strut
      end
      tex.print(': &')
      for j, node_ in ipairs(nodelist_) do
         if node_.empty then
            if node_.phantom then
               tex.print('\\hspace{1.5ex}')
            end
            goto finishcell
         end
         if node_.arrow then
            tex.print('$\\to ' .. node_.base .. '$')
            goto finishcell
         end
         dotter = ''
         if node_.dots > 0 then
            dotter = '\\'
            for _ = 1, node_.dots do
               dotter = dotter .. 'd'
            end
            dotter = dotter .. 'ot '
         end
         super = ''
         if string.len(node_.superscript) > 0 then
            first = string.byte(string.sub(node_.superscript, 1, 1))
            if first >= 65 and first <= 90 then -- only big letters are relevant
               hassuper = true
            end
            super = '^{' .. node_.superscript .. '}'
         end
         sub = ''
         if string.len(node_.subscript) > 0 then
            sub = '_{' .. node_.subscript .. '}'
         end
         tex.print('$' .. dotter .. node_.base .. super .. sub .. '$')
         ::finishcell::
         if j < #nodelist_ then
            tex.print('&')
         end
      end
      if hassuper then
         tex.print('\\vphantom{\\scalebox{1.45}{M}}') -- a strut
      end
      tex.print('\\\\')
   end
   tex.print('\\end{tabular}')
   tex.print('\\bgroup\\makeatletter')
   tex.print('\\protected@write\\@auxout{}{'
             .. '\\string\\gdef\\string\\transitionsequenceforpattern'
             .. _str_to_safe_str(current_pattern) .. '{'
             .. _transition_sequence(jugglers, nodes) ..'}}')
   tex.print('\\egroup')
end

function transition_sequence()
   local cmd = ('\\transitionsequenceforpattern'
                .. _str_to_safe_str(current_pattern))
   tex.print('\\ifdefined' .. cmd .. cmd .. '\\else...\\fi{}')
end

local function _rotate_string(s, offset)
   local prefix, postfix
   prefix = string.sub(s, 1, offset)
   postfix = string.sub(s, offset + 1)
   return postfix .. prefix
end

local function _index_modulo(x, modulus)
   return ((x - 1) % modulus) + 1
end

local function _score_interface(interface)
   -- move non-empty slots to the front giving different weights to
   -- different types to get a well-defined maximum for normalization
   local ret, char
   ret = 0
   for i = 1,#interface do
      ret = 2*ret
      char = string.sub(interface, i, i)
      if char == '&' then
         ret = ret + 1/2
      elseif char == 'o' then
         ret = ret + 1/3
      elseif char == 'x' then
         ret = ret + 1/5
      elseif char == '.' then
         ret = ret + 0
      else
         print("ERROR: Unexpected char " .. char)
      end
   end
   return ret
end

local function _replace_sides(interface)
   local ret, char
   ret = ""
   for i = 1,#interface do
      char = string.sub(interface, i, i)
      if char == '&' then
         ret = ret .. '&'
      elseif char == 'x' then
         ret = ret .. 'o'
      elseif char == 'o' then
         ret = ret .. 'x'
      elseif char == '.' then
         ret = ret .. '.'
      else
         print("Error: Unexpected char " .. char)
      end
   end
   return ret
end

local function _normalize_interface(interface)
   local best_val, best_score, val, score
   best_score = 0
   for i = 10, 2, -1 do
      if #interface % i == 0 then
         chunksize = #interface / i
         reference = string.sub(interface, 1, chunksize)
         redundant = true
         for j = 2,i do
            chunk = string.sub(interface, 1+(j-1)*chunksize, j*chunksize)
            if chunk ~= reference then
               redundant = false
               break
            end
         end
         if redundant then
            interface = reference
         end
      end
   end
   for i = 1,#interface do
      val = _rotate_string(interface, i-1)
      if current_interface_kind == 'async' and i % 2 == 0 then
          val = _replace_sides(val)
      end
      score = _score_interface(val)
      if score > best_score then
         best_score = score
         best_val = val
      end
   end
   return best_val
end

local function parse_base_36(num)
   if num == '0' then
       return 0
   elseif num == '1' then
       return 1
   elseif num == '2' then
       return 2
   elseif num == '3' then
       return 3
   elseif num == '4' then
       return 4
   elseif num == '5' then
       return 5
   elseif num == '6' then
       return 6
   elseif num == '7' then
       return 7
   elseif num == '8' then
       return 8
   elseif num == '9' then
       return 9
   elseif num == 'a' then
       return 10
   elseif num == 'b' then
       return 11
   elseif num == 'c' then
       return 12
   elseif num == 'd' then
       return 13
   elseif num == 'e' then
       return 14
   elseif num == 'f' then
       return 15
   elseif num == 'g' then
       return 16
   elseif num == 'h' then
       return 17
   elseif num == 'i' then
       return 18
   elseif num == 'j' then
       return 19
   elseif num == 'k' then
       return 20
   elseif num == 'l' then
       return 21
   elseif num == 'm' then
       return 22
   elseif num == 'n' then
       return 23
   elseif num == 'o' then
       return 24
   elseif num == 'p' then
       return 25
   elseif num == 'q' then
       return 26
   elseif num == 'r' then
       return 27
   elseif num == 's' then
       return 28
   elseif num == 't' then
       return 29
   elseif num == 'u' then
       return 30
   elseif num == 'v' then
       return 31
   elseif num == 'w' then
       return 32
   elseif num == 'x' then
       return 33
   elseif num == 'y' then
       return 34
   elseif num == 'z' then
       return 35
   else
      print("ERROR: unknown numeral " .. num)
   end
end

local function _siteswap_to_interface(siteswap)
   local catches, ifacelen, instructions, num, offset, ret
   catches = {}
   ifacelen = #siteswap
   instructions = siteswap
   if ifacelen % 2 == 1 then
       ifacelen = 2*ifacelen
       instructions = instructions .. siteswap
   end
   for i = 1,ifacelen do
      catches[i] = false
   end
   for i = 1,#instructions do
      num = parse_base_36(string.sub(instructions, i, i))
      if num % 2 == 1 then
         offset = num - 4
         catches[_index_modulo(i + offset, ifacelen)] = true
      end
   end
   ret = ""
   for i = 1,ifacelen do
      if catches[i] then
         if i % 2 == 1 then
            ret = ret .. 'o'
         else
            ret = ret .. 'x'
         end
      else
         ret = ret .. '.'
      end
   end
   return ret
end

function set_current_pattern(label)
   current_pattern = label
end

function index_interface(rawinterface)
   local kind, payload, interface, temp
   if string.len(rawinterface) > 0 then
      kind = string.sub(rawinterface, 1, 1)
      payload = string.sub(rawinterface, 2)
      if kind == 'S' then
         current_interface_kind = "async" -- export via global variable
         interface = _siteswap_to_interface(payload)
         interface = _normalize_interface(interface)
         current_interface = interface -- export via global variable
      elseif kind == '=' then
         current_interface_kind = "sync" -- export via global variable
         interface = _normalize_interface(payload)
         current_interface = interface -- export via global variable
      elseif kind == '>' then
         current_interface_kind = "async" -- export via global variable
         interface = _normalize_interface(payload)
         current_interface = interface -- export via global variable
      end
      temp = interface_store[current_interface_kind][current_interface]
      if not temp then
         temp = {}
         interface_store[current_interface_kind][current_interface] = temp
      end
      table.insert(temp, current_pattern)
   end
end

local function _interface_repeated(interface, interface_kind)
   local half, front, back
   if interface_kind ~= 'async' or #interface % 2 == 1 then
      return false
   end
   half = #interface/2
   for i = 1,half do
       front = string.sub(interface, i, i)
       back = string.sub(interface, i+half, i+half)
       if not ((front == '.' and back == '.')
               or (front == 'o' and back == 'x')
               or (front == 'x' and back == 'o')) then
           return false
       end
   end
   return true
end

function print_current_interface()
   local interface, interface_kind, repeated
   tex.print('\\poi{interface}{')
   interface = current_interface
   interface_kind = current_interface_kind
   repeated = false
   if _interface_repeated(interface, interface_kind) then
       interface = string.sub(interface, 1, #interface / 2)
       repeated = true
   end
   for i = 1,#interface do
      if string.sub(interface, i, i) == '&' then
         tex.sprint('\\ifaceboth{}\\,{}')
      elseif string.sub(interface, i, i) == 'o' then
         tex.sprint('\\ifaceown{}\\,{}')
      elseif string.sub(interface, i, i) == 'x' then
         tex.sprint('\\ifacepartner{}\\,{}')
      elseif string.sub(interface, i, i) == '.' then
         tex.sprint('\\ifacenop{}\\,{}')
      else
         print("ERROR: unexpected interface character")
      end
   end
   if repeated then
      tex.sprint('\\ifacerepeat{}\\,{}')
   end
   if interface_kind == "sync" then
      tex.print('(sync)')
   elseif interface_kind == "async" then
      tex.print('(async)')
   else
      print('WARNING: unknown interface kind ' .. interface_kind)
   end
   tex.print('}')
end

-- Taken from https://www.lua.org/pil/19.3.html
local function pairsByKeys (t, f)
   local a = {}
   for n in pairs(t) do table.insert(a, n) end
   table.sort(a, f)
   local i = 0      -- iterator variable
   local iter = function ()   -- iterator function
      i = i + 1
      if a[i] == nil then return nil
      else return a[i], t[a[i]]
      end
   end
   return iter
end

function print_interface_index()
   local s
   tex.print('\\begin{itemize}')
   for _, store in pairsByKeys(interface_store) do
      for _, patterns in pairsByKeys(store) do
         if #patterns > 1 then
             tex.print('\\item ')
             for i, label in ipairs(patterns) do
                s = '\\refer{' .. label .. '}'
                if i < #patterns then
                   s = s .. ', '
                end
                tex.print(s)
             end
         end
      end
   end
   tex.print('\\end{itemize}')
end

function initialize_variables()
   interface_store = {sync = {}, async = {}}
   preview_toggle = false
   current_pattern = nil
   -- Use luas global variables to do a bit of magic
   A = 'A'
   B = 'B'
   C = 'C'
   D = 'D'
   E = 'E'
   F = 'F'
   G = 'G'
   H = 'H'
   I = 'I'
   J = 'J'
   K = 'K'
   L = 'L'
   M = 'M'
   N = 'N'
   O = 'O'
   P = 'P'
   Q = 'Q'
   R = 'R'
   S = 'S'
   T = 'T'
   U = 'U'
   V = 'V'
   W = 'W'
   X = 'X'
   Y = 'Y'
   Z = 'Z'
   r = 'r'
   l = 'l'
   rl = 'rl'
   lr = 'lr'
end

function enable_preview()
   preview_toggle = true
end
